/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#define TRUE 1
#define FALSE 0
#define _GNU_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>

/* #define DEBUG */

#ifdef DEBUG
#define D_(fmt,...) printf(fmt, ## __VA_ARGS__)
#else
#define D_(fmt,...)
#endif

/*the string "blah" must be the FIRST parameter, but his improves perf a little bit */
#define MATCH(PATTERN, LINE) (strncmp(PATTERN, skip_spaces(LINE), sizeof(PATTERN) - 1) == 0)
#define LEN(STR) (sizeof(STR) - 1)

static char **path;

/* for the latter paths, a trailing '/' is required ! */
static const char *additional_paths[] = {
    "/bin/",
    "/sbin/",
    "/usr/bin/",
    "/usr/sbin/",
    "/usr/local/bin/",
    "/usr/local/sbin/",
    NULL
};

static char *at_default_path = NULL;

/* #define MAX_AT_DEF_PATHS 100 */
static char **at_default_paths = NULL;

/* this program will parse all .i files in installation
   setting distribution dependent values on install time */

static int _is_file(const char *filename, int executable)
{
    struct stat tmp;

    D_("Trying, with file \"%s\" : ", filename);
    if (stat(filename, &tmp) == 0 && S_ISREG(tmp.st_mode))
    {
        if (executable)
        {
            if (!
                (tmp.
                 st_mode & (S_IXUSR | S_IXGRP | S_IXOTH | S_ISUID | S_ISGID |
                            S_ISVTX)))
            {
                D_("Found it, but not an executable\n");
                return (FALSE);
            }
        }
        D_("Found it.\n");
        return (TRUE);
    }
    D_("Did not find it.\n");
    return (FALSE);
}

#define is_file(FILENAME) _is_file(FILENAME, 0)
#define is_exec(FILENAME) _is_file(FILENAME, 1)

static const char *probe_distribution(void)
{
    D_("probe_distribution\n");
    if (is_file("/etc/debian_version"))
        return ("debian");
    if (is_file("/etc/gentoo-release") || is_file("/etc/pardus-release"))
        return ("gentoo");
    if (is_file("/etc/fedora-release"))
        return ("fedora");
    if (is_file("/etc/pld-release"))
        return ("pld");
    if (is_file("/etc/pingwinek-release"))
        return ("pingwinek");
    if (is_file("/etc/altlinux-release"))
        return ("altlinux");
    D_("Probe failed\n");
    return ("unknown");
}

/* this should split a sring into an "argv" like array using char delim */
/* dont use this any more, due to some obscure memory mgmt bug! 
struct my_ll
{
    char *data;
    struct my_ll *next;
}; */

/* libc provides strtok, but they said, i should not use it */
static char **str_split(const char *in, char delim)
{
    /*struct my_ll *head;
    struct my_ll *current;*/
    int num_tokens = 1;
    char *tmp;
    char **retval;

    /* count the number of tokens. */
    /* there is at least one token plus one for each delim char found */
    /* it doesn't matter if we allocate a bit more mem than required, eg for :: */
    for (tmp = (char *)in; *tmp; tmp++)
        if (*tmp == ':')
            num_tokens++;

    retval = malloc(sizeof(char *) * (num_tokens + 1));

    /* there may be delims in front of in */
    while (*in == delim)
        in++;

    tmp = (*retval = strdup(in));

    num_tokens = 1;

    while ((tmp = strchr(tmp, delim)))
    {
        *tmp = '\0';
        tmp++;
        while (*tmp == delim)
            tmp++;                          /*skip additional delims */
        if (*tmp == '\0')
            break;                          /*in case a delim was at the end of in */
        /* ok, here we found a new token */
        retval[num_tokens++] = tmp;
    }
    retval[num_tokens] = NULL;

    D_("got %d tokens\n", num_tokens);
    return retval;
}

/* This should find string like @mknod@ and return a pointer to it if found,
 * otherwise return NULL */

static const char *find_at_phrases(const char *in)
{
    const char *first;
    const char *second;
    const char *tmp;

    if (!(first = strchr(in, '@')))
        return NULL;                        /* no '@' */
    while (1)
    {
        if (!(second = strchr(first + 1, '@')))
            return NULL;                    /* only one '@' */

        for (tmp = first + 1; tmp < second; tmp++)
            if (*tmp == ' ' || *tmp == '\t')
            {
                /* whitespace in between: no match */
                first = second;
                continue;
            }
        return first;
    }
}

static int parse_phrase_no_colon(FILE * stream, const char *fullphrase,
                                 int def_path_no)
{
#define FILENAME_LEN 1000
    char filename[FILENAME_LEN];
    const char *tmp;
    const char *phrase;
    int i;
    int len;

    D_("ppnc: fp=\"%s\", i=%d\n", fullphrase, def_path_no);
    if (!*fullphrase)
    {
        fputc('@', stream);
        return 0;                           /* @@ => @ translation, no looping required */
    }

    /*    if (!path)
       return; */
    if (!(phrase = strrchr(fullphrase, '/')))
        phrase = fullphrase;
    else
        phrase++;

    if (phrase != fullphrase)
    {
        /* this means fullphrase contains an / */
        if (is_exec(fullphrase))
        {
            fputs(fullphrase, stream);
            return 0;                       /* found filename in supplied Path => no looping req'd */
        }
    }
    /* first, search the path */
    i = 0;
    while (path[i])
    {
        strncpy(filename, path[i], FILENAME_LEN - 2);
        len = strlen(filename);
        if (filename[len - 1] != '/')
        {
            filename[len] = '/';
            filename[len + 1] = '\0';
        }
        strncat(filename, phrase, FILENAME_LEN - 1 - strlen(filename));
        if (is_exec(filename))
        {
            fputs(filename, stream);
            return 0;                       /* found filename in Path => no looping req'd */
        }
        i++;
    }
    /* then, search additional paths */
    i = 0;
    while (additional_paths[i])
    {
        strncpy(filename, additional_paths[i], FILENAME_LEN - 1);
        strncat(filename, phrase,
                FILENAME_LEN - 1 - strlen(additional_paths[i]));
        if (is_exec(filename))
        {
            fputs(filename, stream);
            return 0;                       /* found filename in Path => no looping req'd */
        }
        i++;
    }
    /* here we neither found it in the path, nor in the additional paths */
    if (def_path_no == -1)
        return 1;                           /* this is from a coloned search: the first bastard wasn't found, try the next */
    fprintf(stderr, "WARNING: No executable found for \"%s\", ", phrase);

    if (fullphrase != phrase)
    {
        fprintf(stderr, "using full path \"%s\"\n", fullphrase);
        fputs(fullphrase, stream);
        /* return 0;  full path specified => no looping req'd */
    }
    else if (at_default_path)
    {
        tmp = at_default_paths[def_path_no];
        fprintf(stderr, "using supplied default path \"%s\"\n", tmp);
        fputs(tmp, stream);
        if (tmp[strlen(tmp) - 1] != '/')
            fputc('/', stream);
        fputs(phrase, stream);
        if (at_default_paths[def_path_no + 1])
            return 1;                       /* next default path exits, so looping is required */
        /* return 0 */
    }
    else                                    /* at_default_path == NULL && fullphrase == phrase */
    {
        fprintf(stderr, "using builtin default path \"/usr/sbin/\"\n");
        fputs("/usr/sbin/", stream);
        fputs(phrase, stream);
        /* no default path(s) so no looping */
    }
    /* exit(1) ??? */
    return 0;



#if 0
    tmp = path;
    while (1)
    {
        i = 0;
        while (*tmp != ':' && *tmp != '\0' && i < FILENAME_LEN - 1)
            filename[i++] = *(tmp++);
        if (*tmp)
            tmp++;                          /* skip ':' */
        if (filename[i] != '/')
            filename[i++] = '/';
        filename[i] = '\0';
        strncat(filename, phrase, FILENAME_LEN - 1 - i);
        /* TODO: check if executable */
        if (is_exec(filename))
        {
            fputs(filename, stream);
            return 0;                       /* found filename in Path => no looping req'd */
        }
        if (!*tmp)
        {
            /* here we passed all paths in PATH */
            i = 0;
            while (additional_paths[i])
            {
                strncpy(filename, additional_paths[i], FILENAME_LEN - 1);
                strncat(filename, phrase,
                        FILENAME_LEN - 1 - strlen(additional_paths[i]));
                if (is_exec(filename))
                {
                    fputs(filename, stream);
                    return 0;               /* found filename in Path => no looping req'd */
                }
                i++;
            }
            if (def_path_no == -1)
                return 1;                   /* this is from a coloned search: the first bastard wasn't found, try the next */
            fprintf(stderr, "WARNING: No executable found for \"%s\", ",
                    phrase);
            if (fullphrase != phrase)
            {
                fprintf(stderr, "using full path \"%s\"\n", fullphrase);
                fputs(fullphrase, stream);
                /* return 0;  full path specified => no looping req'd */
            }
            else if (at_default_path)
            {
                tmp = at_default_paths[def_path_no];
                fprintf(stderr, "using supplied default path \"%s\"\n", tmp);
                fputs(tmp, stream);
                if (tmp[strlen(tmp) - 1] != '/')
                    fputc('/', stream);
                fputs(phrase, stream);
                if (at_default_paths[def_path_no + 1])
                    return 1;               /* next default path exits, so looping is required */
                /* return 0 */
            }
            else                            /* at_default_path == NULL && fullphrase == phrase */
            {
                fprintf(stderr,
                        "using builtin default path \"/usr/sbin/\"\n");
                fputs("/usr/sbin/", stream);
                fputs(phrase, stream);
                /* no default path(s) so no looping */
            }
            /* exit(1) ??? */
            return 0;
        }
    }
#endif
}

static int parse_phrase(FILE * stream, const char *fullphrase,
                        int def_path_no)
{
    char **phrases;

    /* char* tmp; */
    int i = 0;
    int rv;

    D_("pp: i=%d; called with %s\n", def_path_no, fullphrase);
#if 0
    phrases[0] = strdup(fullphrase);

    while ((tmp = strchr(phrases[i], ':')))
    {
        *tmp = '\0';
        phrases[++i] = tmp + 1;
        if (i == 98)
            break;
    }
    phrases[i + 1] = NULL;
#endif
    phrases = str_split(fullphrase, ':');
    if (def_path_no == 0)
    {
        for (i = 0; phrases[i]; i++)
        {
            D_("doing ppnc(%s)\n", phrases[i]);
            if (parse_phrase_no_colon(stream, phrases[i], -1) == 0)
            {
                free(phrases[0]);
                return 0;                   /* found it */
            }
        }
        /*xxx */
        /* here = bad: all executables have not been found ! */
    }

    rv = parse_phrase_no_colon(stream, phrases[0], def_path_no);
    free(phrases[0]);
    free(phrases);
    return rv;
}

static int print_line(FILE * stream, const char *buf, int def_path_no)
{
    const char *tmp;
    const char *phrase;
    int i = 0;
    int retval = 0;

#define PHRASE_LEN 1000
    char phrase_name[PHRASE_LEN];

    while (1)
    {
        phrase_name[0] = '\0';
        if (!(phrase = find_at_phrases(buf)))
        {
            fputs(buf, stream);
            return retval;
        }

        /* write all chars up to the beginning '@' thingie to stream */
        for (tmp = buf; tmp < phrase; tmp++)
            fputc(*tmp, stream);

        /* copy Phrase to phrase_name */
        i = 0;
        for (tmp = phrase + 1; *tmp != '@' && i < PHRASE_LEN; tmp++)
        {
            phrase_name[i++] = *tmp;
        }
        phrase_name[i] = '\0';
        /* pasrse Phrase */
        retval |= parse_phrase(stream, phrase_name, def_path_no);
        /* re-adjust output buffer */
        buf = tmp + 1;
    }
}

#define do_print_line(STREAM, BUF) for(i = 0; print_line(STREAM, BUF, i); i++)

static void print_it(FILE * stream, const char *buf)
{
    char *line, *tmp;
    int i;

    D_("print_it: Got block \"%s\"\n", buf);
    while (1)
    {
        tmp = strchr(buf, '\n') + 1;
        if (!tmp)
        {
            do_print_line(stream, buf);
            return;
        }
        else
        {
            line = strndup(buf, tmp - buf);
            buf = tmp;
            D_("print_it: Got line \"%s\"\n", line);
            do_print_line(stream, line);
            free(line);
        }
        if (!*buf)
            return;
    }
}

static char *do_exec(char *data, int length)
{
    int pipe_fork[2], inpipe_fork[2];

    pipe(pipe_fork);
    pipe(inpipe_fork);

    int status;
    int datalen = 0;
    int pid;
    char *return_data = NULL;

    D_("forking...\n");
    if ((pid = fork()) == 0)
    {
        const char *argv[] = { "/bin/bash", NULL, NULL };
        dup2(pipe_fork[0], 0);
        close(pipe_fork[1]);
        dup2(inpipe_fork[1], 1);
        dup2(inpipe_fork[1], 2);
        close(inpipe_fork[0]);
        execve("/bin/bash", (char **) argv, environ);
        printf(" ** execve /bin/bash FAILED!\n");
        exit(1);
    }
    else
    {
        D_("This is not fork\n");
        return_data = malloc(1024);
        close(pipe_fork[0]);

        /*        printf("Writing data to pipe : (%d) \n",length);
           int x;
           for (x = 0; x <= length;x++)
           printf("%c",data[x]);
           printf("\n"); */

        D_("Data sent to fork (%i): \"%s\"\n", length, data);
        write(pipe_fork[1], data, length);
        write(pipe_fork[1], "\n\nexit\n", 7);
        close(pipe_fork[1]);

        wait(&status);

        int fd_flags;

        fd_flags = fcntl(inpipe_fork[0], F_GETFL, 0);
        fcntl(inpipe_fork[0], F_SETFL, fd_flags | O_NONBLOCK);
        close(inpipe_fork[1]);
        D_("Reading data from fork\n");
        datalen = read(inpipe_fork[0], return_data, 1023);
        close(inpipe_fork[0]);
        if (datalen > 0)
        {
            D_("Read %d bytes: \"%s\"\n", datalen, return_data);
            return_data[datalen] = '\0';
            return return_data;
        }
        printf("Did not got any data from #exec statement!\n");
        free(return_data);
        return NULL;
    }
    return NULL;
}

static inline const char *skip_spaces(const char *in)
{
    while (*in == ' ' || *in == '\t')
        in++;
    return in;
}

static inline int empty_elsed(const char *in)
{
    in = skip_spaces(in);
    in += LEN("#elsed");
    while (*in == ' ' || *in == '\t')
        in++;
    if (*in == '\n')
        return 1;
    else
        return 0;
}

/* this verifies that the string "distro" is found in the "in" string
 * and that the character before the string "distro" in "in" is either
 * ' ' or '\t'
 * the character after "distro" meight either be ' ', '\t' or '\n'
 */
static inline int match_distro(const char *in, const char *distro)
{
    const char *tmp;
    int len = strlen(distro);

    tmp = skip_spaces(in);
    while (1)
    {
        tmp = strstr(tmp, distro);
        if (!tmp)
            return 0;
        if ((tmp[-1] != ' ' && tmp[-1] != '\t')
            || (tmp[len] != ' ' && tmp[len] != '\t' && tmp[len] != '\n'))
            tmp++;                          /* this will make strstr not find the same part again */
        else
            return 1;
    }
}

int main(int argc, char **argv)
{
    int i = 1;

    FILE *in;
    FILE *out = NULL;

    const char *distro = NULL;
    const char *infile = NULL;
    const char *outfile = NULL;
    const char *data;

    /* char* tmp; */

#define LINE_LEN 1000
    char line[LINE_LEN];

#define EXEC_BUFFER_LEN 100000
    char exec_buffer[EXEC_BUFFER_LEN];
    int exec_buffer_free = EXEC_BUFFER_LEN - 1;

#define IF_BLOCK 1
#define EXEC_BLOCK 2
    int in_block = 0;
    int if_block_processed = 0;
    int if_block_print_out = 0;

    path = str_split(getenv("PATH"), ':');

    if (!path)
    {
        fprintf(stderr, "ERROR: could not get $PATH\n");
        exit(1);
    }
    /* check cmdline for set distribution */
    while (argv[i])
    {

        if (argv[i + 1])
        {
            if (strcmp(argv[i], "--dist") == 0 || strcmp(argv[i], "-d") == 0)
            {
                distro = argv[i + 1];
                i++;
            }
            else if (strcmp(argv[i], "--in") == 0
                     || strcmp(argv[i], "-i") == 0)
            {
                infile = argv[i + 1];
                i++;
            }
            else if (strcmp(argv[i], "--out") == 0
                     || strcmp(argv[i], "-o") == 0)
            {
                outfile = argv[i + 1];
                i++;
            }
        }
        i++;
    }

    /* if not set, probe it */
    if (!distro)
        distro = probe_distribution();
    if (!distro)
    {
        fprintf(stderr, "Unknown DISTRIBUTION\n");
        exit(1);
    }
    fprintf(stderr, "Distribution set: \"%s\"\n", distro);

    /* open input for writing */
    if (!infile)
    {
        fprintf(stderr, "You have to set input file with -i\n");
        exit(1);
    }
    if (!(in = fopen(infile, "r")))
    {
        fprintf(stderr, "could not open input file!\n");
        exit(1);
    }

    /* open output for writing */
    if (!outfile)
    {
        if (strstr(infile, ".ii"))
        {
            outfile = strndup(infile, strlen(infile) - 1);
        }
        else
        {
            fprintf(stderr, "out with -o not set, defaulting to stdout.\n");
            out = stdout;
        }
    }
    if (!out && !(out = fopen(outfile, "w")))
    {
        fprintf(stderr, "could not open output file!\n");
        exit(1);
    }

    /* main loop */
    while (1)
    {
        fgets(line, LINE_LEN, in);
        if (feof(in))
            break;
        if (MATCH("#atdefpath", line))
        {
            //char* tmp;
            D_("found #atdefpath. updating at default path\n");
            data = skip_spaces(line + LEN("#atdefpath"));
            if (at_default_path)
                free(at_default_path);
            at_default_path = malloc(strlen(data) + 1);
            i = 0;
            while (data[i] && data[i] != ' ' && data[i] != '\t'
                   && data[i] != '\n')
            {
                at_default_path[i] = data[i];
                i++;
            }
            /*            if (at_default_path[i - 1] != '/')
               at_default_path[i++] = '/'; */
            at_default_path[i] = '\0';
            D_("at_default_path updated to %s\n", at_default_path);

            at_default_paths = str_split(at_default_path, ':');
#if 0
            at_default_paths[0] = at_default_path;

            i = 0;
            while ((tmp = strchr(at_default_paths[i], ':')))
            {
                *tmp = '\0';
                i++;
                at_default_paths[i] = tmp + 1;
                if (i == MAX_AT_DEF_PATHS - 1)
                    break;                  /* inner loop */
            }
            at_default_paths[i + 1] = NULL;
#endif
            continue;
        }

        if (!in_block && line[0] != '#')
        {
            print_it(out, line);
            continue;                       /* read next line */
        }

        /* either in_block != 0 or line started with '#' */
        if (!in_block)
        {
            if (MATCH("#exec", line))
            {
                D_("found #exec \n");
                exec_buffer[0] = '\0';
                exec_buffer_free = EXEC_BUFFER_LEN - 1;
                in_block = EXEC_BLOCK;
                continue;
            }
            else if (MATCH("#ifd", line))
            {
                D_("found #ifd\n");
                in_block = IF_BLOCK;
                if_block_print_out = 0;
                if_block_processed = 0;
                if (match_distro(line, distro))
                {
                    D_("actual distro %s matches line %s", distro, line);
                    if_block_print_out = 1;
                    if_block_processed = 1;
                }
                continue;
            }
            else
            {
                /* line started with a '#', but we are not in_block, and the line was neither 
                 * #ifd nor #exec. so it's a comment.
                 */
                if (MATCH("#elsed", line) ||
                    MATCH("#endd", line) || MATCH("#endexec", line))
                {
                    fprintf(stderr,
                            "WARNING: found #elsed, #end or #endexec, but not in BLOCK!\n");
                    /* exit here ??? */
                }
                /* this is a comment so print it */
                print_it(out, line);
                continue;                   /* read next line */
            }
        }
        /* ok, we are in a block */
        if (in_block == IF_BLOCK)
        {
            if (if_block_print_out)
            {
                /*if the block should be printed up to now */
                if (MATCH("#endd", line) || MATCH("#elsed", line))
                {
                    /* elsed or endd, but this section should be printed => stop printing */
                    if_block_print_out = 0;
                    /* endd => in_block = 0; */
                    if (MATCH("#endd", line))
                    {
                        D_("found #endd\n");
                        in_block = 0;
                    }
                    else
                    {
                        D_("found #elsed\n");
                    }
                    continue;
                }
                else
                {
                    /* in if block that should be printed => print it */
                    print_it(out, line);
                    continue;
                }

            }
            else
            {
                /* if block that should not be printed: check for elsed, endd */
                if (MATCH("#elsed", line))
                {
                    D_("found #elsed\n");
                    /* now we need to check if there is some distro specified */
                    if (empty_elsed(line))
                    {
                        /*no distro specified */
                        if (!if_block_processed)
                        {
                            /* this if block has not yet been processed, so print this one */
                            if_block_print_out = 1;
                        }
                        continue;
                    }
                    else
                    {                       /* #elsed blahblah */
                        if (match_distro(line, distro))
                        {
                            if_block_print_out = 1;
                            if_block_processed = 1;
                            continue;
                        }
                    }
                }
                else if (MATCH("#endd", line))
                {
                    D_("found #endd\n");
                    in_block = 0;
                    continue;
                }
                else
                    continue;
            }
        }
        else
        {
            /* in_block == EXEC_BLOCK */
            if (MATCH("#endexec", line))
            {
                D_("found #endexec\n");
                /* Do the exec shit */
                data = do_exec(exec_buffer, strlen(exec_buffer));
                if (data)
                {
                    print_it(out, data);
                    free((void *) data);
                }
                in_block = 0;
                continue;
            }
            else
            {
                strncat(exec_buffer, line, exec_buffer_free);
                exec_buffer_free -= strlen(line);
                continue;
            }
        }
    }
    fclose(in);
    fclose(out);
    exit(0);
}
